# This file is modified from the screenutils Python module
# https://pypi.org/project/screenutils/
# https://github.com/Christophe31/screenutils

# -*- coding:utf-8 -*-
#
# This program is free software. It comes without any warranty, to
# the extent permitted by applicable law. You can redistribute it
# and/or modify it under the terms of the GNU Public License 2 or upper.
# Please ask if you wish a more permissive license.

try:
    from commands import getoutput
except Exception:
    from subprocess import getoutput
from os import system
from time import sleep


class ScreenNotFoundError(Exception):
    """Raised when the screen does not exists."""

    def __init__(self, message, screen_name):
        message += " Screen \"{0}\" not found".format(screen_name)
        self.screen_name = screen_name
        super(ScreenNotFoundError, self).__init__(message)


def list_screens():
    """List all the existing screens and build a Screen instance for each."""
    list_cmd = "screen -ls"
    return [
        Screen(".".join(l.split(".")[1:]).split("\t")[0])
        for l in getoutput(list_cmd).split('\n')
        if "\t" in l and ".".join(l.split(".")[1:]).split("\t")[0]
    ]


class Screen(object):
    """Represents a gnu-screen object.

    >>> s=Screen("screenName", initialize=True)
    >>> s.name
    'screenName'
    >>> s.exists
    True
    >>> s.state
    >>> s.send_commands("man -k keyboard")
    >>> s.kill()
    >>> s.exists
    False
    """

    def __init__(self, name, initialize=False):
        self.name = name
        self._id = None
        self._status = None
        if initialize:
            self.initialize()

    @property
    def id(self):
        """Return the identifier of the screen as string."""
        if not self._id:
            self._set_screen_infos()
        return self._id

    @property
    def status(self):
        """Return the status of the screen as string."""
        self._set_screen_infos()
        return self._status

    @property
    def exists(self):
        """Tell if the screen session exists or not."""
        # Parse the screen -ls call, to find if the screen exists or not.
        #  "    28062.G.Terminal    (Detached)"
        lines = getoutput("screen -ls").split('\n')
        return self.name in [
            ".".join(l.split(".")[1:]).split("\t")[0]
            for l in lines
            if self.name in l
        ]

    def initialize(self):
        """Initialize a screen, if does not exists yet."""
        if not self.exists:
            self._id = None
            # Detach the screen once attached, on a new tread.
            # support Unicode (-U),
            # attach to a new/existing named screen (-R).

            # ORIGINAL
            # Thread(target=self._delayed_detach).start()
            # system('screen -s sh -UR -S ' + self.name)

            # CUSTOM
            system('screen -d -m -S ' + self.name)

    def interrupt(self):
        """Insert CTRL+C in the screen session."""
        self._screen_commands("eval \"stuff \\003\"")

    def kill(self):
        """Kill the screen applications then close the screen."""
        self._screen_commands('quit')

    def detach(self):
        """Detach the screen."""
        self._check_exists()
        system("screen -d " + self.id)

    def send_commands(self, *commands):
        """Send commands to the active gnu-screen."""
        self._check_exists()
        for command in commands:

            # use single quote unless that is a part of the command
            if "'" in command:
                q = "\""
            else:
                q = "\'"

            self._screen_commands(
                'stuff {q}{c}{q}'.format(q=q, c=command),
                'eval "stuff \\015"'
            )

    def add_user_access(self, unix_user_name):
        """Allow to share your session with an other unix user."""
        self._screen_commands('multiuser on', 'acladd ' + unix_user_name)

    def _screen_commands(self, *commands):
        """Allow to insert generic screen specific commands."""
        self._check_exists()
        for command in commands:
            cmd = 'screen -x {0}.{1} -p 0 -X {2}'.format(self.id, self.name, command)
            system(cmd)
            sleep(0.02)

    def _check_exists(self, message="Error code: 404."):
        """Check whereas the screen exist. if not, raise an exception."""
        if not self.exists:
            raise ScreenNotFoundError(message, self.name)

    def _set_screen_infos(self):
        """Set the screen information related parameters."""
        if self.exists:
            line = ""
            for l in getoutput("screen -ls").split("\n"):
                if (
                    l.startswith('\t') and
                    self.name in l and
                    self.name == ".".join(l.split('\t')[1].split('.')[1:]) in l
                ):
                    line = l
            if not line:
                raise ScreenNotFoundError("While getting info.", self.name)
            infos = line.split('\t')[1:]
            self._id = infos[0].split('.')[0]
            if len(infos) == 3:
                self._date = infos[1][1:-1]
                self._status = infos[2][1:-1]
            else:
                self._status = infos[1][1:-1]

    def _delayed_detach(self):
        sleep(0.5)
        self.detach()

    def __repr__(self):
        return "<%s '%s'>" % (self.__class__.__name__, self.name)



from sys import version_info as _version_info

if isinstance(_version_info, tuple):
    version = _version_info[0]
else:
    version = _version_info.major

# different importing for python 2 and 3
if version == 2:
    from jarjar import jarjar
    from _version import __version__
    from screen import Screen
else:
    from jarjar.jarjar import jarjar
    from jarjar._version import __version__
    from jarjar.screen import Screen

__all__ = [
    '__version__',
    'Screen',
    'jarjar',
]



"""Class file for jarjar
"""

import requests
import json
import time
import os
import imp
import warnings
import copy


class jarjar(object):
    """A jarjar slack messenger.

    This is largely a wrapper around functionality in ``requests.post()`` with
    facilities to store and set default values for the desired message to
    send, channel to post within, and slack team webhook.

    Inference for these values proceeds as follows.

    1. Any argument provided to :func:`~jarjar.jarjar.text` or
       :func:`~jarjar.jarjar.attach` supersedes all defaults.
    2. Defaults can be provided at initialization or via a config file
       (``~/.jarjar``), which looks like:

    .. code::

        channel="@username"
        message="Custom message"
        webhook="https://hooks.slack.com/services/your/teams/webhook"

    3. Arguments provided  at initialization supersede those in ``~/.jarjar``.
       If the channel or webhook arguments are never provided, an error is
       raised. If the channel and webhook are provided but not a message or
       attachment, jarjar will make something up.

    Methods
    -------
    attach(attach, channel=None, webhook=None, message=None)
        Send an attachment. User may also include a text message.
    text(message, channel=None, webhook=None, attach=None)
        Send a text message. User may also include an attachment.
    set_webhook(webhook)
        Set jarjar's default webhook.
    set_channel(channel)
        Set jarjar's default channel.
    set_message(message)
        Set jarjar's default message.

    Parameters
    ----------
    message : str
        Optional. Default message to send.
    channel : str,  list
        Optional. Name of the default channel to post within.
    webhook : str
        Optional. Webhook URL for the default slack team.

    """

    _expected_kwargs = ['message', 'attach', 'channel', 'webhook']
    _final_default_message = 'Meesa Jarjar Binks!'
    _no_message_warn = (
        '''
        Slow down cowboy! You didn't provide a message and there is
        no default in your .jarjar, so I'll just wing it.
        '''
        .strip()
        .replace('\n', ' ')
        .replace('\t', ' ')
        .replace('  ', ' ')
    )

    # defaults; exposed for the user
    headers = {'Content-Type': 'application/json'}

    def __init__(self, message=None, channel=None, webhook=None):

        # read config file, set defaults
        self._read_config()
        self._set_defaults(channel=channel, webhook=webhook, message=message)

        # default attach and payload args
        self.attachment_args = dict(
            fallback="New attachments are ready!",
            color="#36a64f",
            fields=[]
        )
        self.payload_args = dict()

    def _set_defaults(self, channel=None, webhook=None, message=None):
        """Set the default channel and webhook and message."""
        # set default channel
        if channel in (None, ''):
            self.default_channel = self.cfg_channel
        else:
            self.default_channel = channel

        if webhook in (None, ''):
            self.default_webhook = self.cfg_webhook
        else:
            self.default_webhook = webhook

        if message in (None, ''):
            self.default_message = self.cfg_message
        else:
            self.default_message = message

    def _read_config(self):
        """Read the .jarjar file for defaults."""
        # get .jarjar path
        filename = os.path.join(os.path.expanduser('~'), '.jarjar')

        # make empty .jarjar if needed
        if not os.path.exists(filename):
            open(filename, 'a').close()

        # load config
        cfg = imp.load_source('_jarjar', filename)

        # assign variables
        for field in ['channel', 'webhook', 'message']:

            # read from config, or set to none
            if hasattr(cfg, field):
                data = getattr(cfg, field)
            else:
                data = None

            # set value
            setattr(self, 'cfg_%s' % field, data)

    def _infer_kwargs(self, **kwargs):
        """Infer kwargs for later method calls."""
        def _get(arg):
            """Return provided arg if it exists. Otherwise, infer."""
            if arg in kwargs and kwargs[arg] not in ('', None):
                return kwargs[arg]

            # No support for default attach ATM.
            if arg == 'attach':
                return None

            # get a default
            default = getattr(self, 'default_{}'.format(arg))

            # return defaults for channel and webhook
            if arg in ['channel', 'webhook']:
                if not default:
                    raise NameError('No {} provided!'.format(arg))
                else:
                    return default

            # return default message if provided
            if self.default_message is not None:
                return self.default_message

            # no message is allowed if there is an attach
            if 'attach' in kwargs and kwargs['attach']:
                return None

            # otherwise use a super-default and warn the user.
            warnings.warn(self._no_message_warn)
            return self._final_default_message

        # check unexpected args
        for k, _ in kwargs.items():
            if k in self._expected_kwargs:
                continue
            warnings.warn('Recieved unexpected kwarg: `%s`.' % k)

        result = dict()
        for arg in ['message', 'attach', 'channel', 'webhook']:
            result[arg] = _get(arg)
        return result

    def _attachment_formatter(self, attach):
        """Format a dict to become a slack attachment."""
        attachments = copy.deepcopy(self.attachment_args)
        attachments['ts'] = time.time()

        for key in attach:

            if isinstance(attach[key], str):
                outval = attach[key]
            else:
                try:
                    outval = str(attach[key])
                except UnicodeEncodeError:
                    outval = unicode(attach[key])
                except Exception:
                    raise

            attachments['fields'].append(dict(
                title=key,
                value=outval,
                short=len(outval) < 20
            ))
        return [attachments]

    def attach(self, attach=None, **kwargs):
        """Send an attachment.

        This method is largely identical to :func:`~jarjar.jarjar.text`,
        only differing in the first argument (``attach``), which is expected
        to be a dictionary.

        Parameters
        ----------
        attach : dict
            Attachment data. Optional *but weird if you don't provide one*.
            All values are converted to string for the slack payload so don't
            sweat it.
        message : str
            Text to send. Optional. If attach is None and there is no
            default *and* you don't provide one here, jarjar just wings it.
        channel : str,  list
            Optional. Name of the channel to post within.
            Can also be a list of channel names; jarjar will post to each.
        webhook : str
            Optional. Webhook URL for the slack team.

        Returns
        -------
        response : requests.models.Response
            Requests response object for the POST request to slack.

        """
        if attach is None:
            warnings.warn(
                'You called `attach` but there is no attachment? Weird.'
            )
        kwargs = self._infer_kwargs(attach=attach, **kwargs)
        return self.text(**kwargs)

    def text(self, message=None, **kwargs):
        """Send a text message.

        This method is largely identical to :func:`~jarjar.jarjar.attach`, only
        differing in the first argument (``message``), which is expected to be
        a string.

        Parameters
        ----------
        message : str
            Text to send. Optional *but weird if you don't provide one*.
            If attach is None and there is no default *and* you don't provide
            one here, jarjar just wings it.
        attach : dict
            Attachment data. Optional. All values are converted to string for
            the slack payload so don't sweat it.
        channel : str,  list
            Optional. Name of the channel to post within.
            Can also be a list of channel names; jarjar will post to each.
        webhook : str
            Optional. Webhook URL for the slack team.

        Returns
        -------
        response : requests.models.Response
            Requests response object for the POST request to slack.

        """
        kwargs = self._infer_kwargs(message=message, **kwargs)
        return self._post(**kwargs)

    def post(self, *args, **kwargs):
        """**DEPRECATED**. Pass args to .text and raises a warning."""
        warnings.warn(
            'jarjar.post() is deprecated! I\'ll let this slide ' +
            'but you should switch to text or attach'
        )
        return self.text(*args, **kwargs)

    def _post(self, message=None, attach=None, channel=None, webhook=None):
        """Send a message to slack.

        Arguments are not inferred and all must be provided. Use the `text` or
        `attach` methods for argument inference.

        Parameters
        ----------
        message : str, NoneType
            Text to send.

        attach : dict, NoneType
            Attachment data.

        channel : str,  list
            Optional. Name of the channel to post within.
            Can also be a list of channel names; jarjar will post to each.
        webhook : str
            Optional. Webhook URL for the slack team.

        Returns
        -------
        response : requests.models.Response
            Requests response object for the POST request to slack.

        """
        def _check_arg(arg, name, types, noneable=False):
            """Ensure arguments are valid."""
            # NoneType handler
            if arg is None:
                if not noneable:
                    raise NameError(
                        'User did not provide kwarg `{}`.'.format(name)
                    )
                else:
                    return

            if not isinstance(arg, types):
                raise TypeError(
                    'Kwarg `{0}` has invalid type. Options: ({1})'
                    .format(name, ','.join(map(str, types)))
                )

        # ensure message or attach is provided
        if message is None and attach is None:
            raise NameError('user must provide a message or attachment.')

        # define tupe of string types
        # try/except is control for python 3.
        try:
            str_types = (str, unicode)
        except Exception:
            str_types = (str,)

        # check kwargs
        _check_arg(message, 'message', str_types, noneable=True)
        _check_arg(attach, 'attach', (dict,), noneable=True)
        _check_arg(channel, 'channel', str_types + (list, ))
        _check_arg(webhook, 'webhook', str_types)

        # recursively post to all channels in array of channels
        if isinstance(channel, list):
            status = []
            for c in channel:
                status.append(
                    self._post(
                        message=message,
                        attach=attach,
                        channel=c,
                        webhook=webhook
                    )
                )
            return status

        # construct a payload
        payload = copy.deepcopy(self.payload_args)
        payload['channel'] = channel

        # add text and attachments if provided
        if message is not None:
            payload['text'] = message

        if attach is not None:
            payload['attachments'] = self._attachment_formatter(attach)

        # convert payload to json and return
        payload = json.dumps(payload)
        return requests.post(webhook, data=payload, headers=self.headers)

    def set_webhook(self, webhook):
        """Set default webhook.

        Parameters
        ----------
        webhook : str
            Webhook URL for the slack team.

        """
        self.default_webhook = webhook

    def set_channel(self, channel):
        """Set default channel.

        Parameters
        ----------
        channel : str
            Name of the channel to post within.

        """
        self.default_channel = channel

    def set_message(self, message):
        """Set default message.

        Parameters
        ----------
        message : str
            Default message to send.

        """
        self.default_message = message



from setuptools import setup
from setuptools import find_packages
import re


# get jarjar version
# https://stackoverflow.com/a/7071358/353278
VERSIONFILE = "jarjar/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

long_description = '\
Jarjar is a python utility that makes it easy to send slack notifications \
to your teams. You can import it as a python module or use our command \
line tool.'

# install jarjar
setup(
    name='jarjar',
    version=verstr,
    description='Use python to send messages to your slack team',
    long_description=long_description,
    url='https://github.com/AusterweilLab/jarjar',
    author='The Austerweil Lab at UW-Madison',
    author_email='austerweil.lab@gmail.com',
    license='MIT',
    keywords=['slack', 'messaging'],
    packages=find_packages('.'),
    install_requires=['requests'],
    python_requires='>=2.7,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*',
    zip_safe=False,
    classifiers=[
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    scripts=['bin/jarjar'],
)
